/* -*-c-*-
 *
 * $Id: mp.h,v 1.1 1999/09/03 08:41:12 mdw Exp $
 *
 * Multiprecision arithmetic
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 * 
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: mp.h,v $
 * Revision 1.1  1999/09/03 08:41:12  mdw
 * Initial import.
 *
 */

#ifndef MP_H
#define MP_H

#ifdef __cplusplus
  extern "C" {
#endif

/*----- Header files ------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>

#ifndef MPTYPES_H
#  include "mptypes.h"
#endif

#ifndef MPX_H
#  include "mpx.h"
#endif

#ifndef MPSCAN_H
#  include "mpscan.h"
#endif

/*----- Data structures ---------------------------------------------------*/

typedef struct mp {
  mpw *v;				/* Vector of words */
  unsigned f;				/* Various flags */
  size_t sz;				/* Size allocated to word vector */
  size_t len;				/* Length of current word vector */
} mp;

enum {
  MPF_SIGN = 1,				/* Sign bit */
  MPF_BURN = 2				/* Burn word vector after use */
};

typedef struct mp_bitscan {
  const mp *x;				/* Pointer to target MP */
  size_t i;				/* Index into MP vector */
  mpw w;				/* Current value being examined */
  unsigned bits;			/* Number of bits left in @w@ */
} mp_bitscan;

/*----- External variables ------------------------------------------------*/

extern mp mp_std;

#define MP_ZERO (mp_std + 0)
#define MP_ONE (mp_std + 1)
#define MP_TWO (mp_std + 2)
#define MP_THREE (mp_std + 3)
#define MP_FOUR (mp_std + 4)
#define MP_FIVE (mp_std + 5)
#define MP_TEN (mp_std + 6)
#define MP_MONE (mp_std + 7)

/*----- Memory allocation and low-level fiddling --------------------------*/

/* --- @mp_create@ --- *
 *
 * Arguments	@mp *x@ = pointer to MP head
 *
 * Returns:	---
 *
 * Use:		Initializes an MP ready for use.  The initial value is zero.
 */

#define MP_INIT { 0, 0, 0, 0 }

extern void mp_create(mp */*x*/);

/* --- @MP_BURN@ --- *
 *
 * Arguments:	@x@ = pointer to MP head
 *
 * Use:		Burns the contents of the MP, if it contains sensitive data.
 */

#define MP_BURN(x) do {							\
  mp *_y = (x)								\
  if (_y->v && _y->f & mpf_burn) {					\
    memset(_y->v, 0, _y->sz * sizeof(mpw));				\
    _y->f &= ~MPF_BURN;							\
  }									\
} while (0)

/* --- @mp_free@ --- *
 *
 * Arguments:	@mp *x@ = pointer to MP head
 *
 * Returns:	---
 *
 * Use:		Releases the memory used by an MP.
 */

#define MP_DESTROY(x) do {						\
  mp *_x = (x);								\
  MP_BURN(_x);								\
  if (_x->v)								\
    free(_x->v);							\
  _x->v = 0;								\
  _x->f = 0;								\
  _x->sz = 0;								\
  _x->len = 0;								\
} while (0)  
		   
extern void mp_free(mp */*x*/);

/* --- @MP_ENSURE@ --- *
 *
 * Arguments:	@x@ = pointer to MP head
 *		@len@ = length required (in words)
 *
 * Use:		Ensures that the MP has enough memory to store a @len@-word
 *		value.
 */

#define MP_ENSURE(x, len) do {						\
  mp *_x = (x);								\
  size_t _len = (len);							\
  if (_x->sz < _len)							\
    mp_resize(_x, _len);						\
} while (0)

/* --- @mp_resize@ --- *
 *
 * Arguments:	@mp *x@ = pointer to MP head
 *		@size_t sz@ = size required (in words)
 *
 * Returns:	---
 *
 * Use:		Resizes the MP so that its word vector has space for
 *		exactly @sz@ words.
 */

extern void mp_resize(mp */*x*/, size_t /*sz*/);

/* --- @mp_norm@ --- *
 *
 * Arguments:	@mp *x@ = pointer to MP head
 *
 * Returns:	---
 *
 * Use:		`Normalizes' an MP.  Fixes the @len@ field so that it's
 *		correct.  Assumes that @len@ is either already correct or
 *		too large.
 */

#define MP_NORM(x) do {							\
  mp *_y = (x);								\
  MPX_LEN(_y->len, _y->v, _y->len);					\
} while (0)

extern void mp_norm(mp */*x*/);

/* --- @mp_dump@ --- *
 *
 * Arguments:	@mp *x@ = pointer to MP head
 *		@FILE *fp@ = pointer to stream to write on
 *
 * Returns:	---
 *
 * Use:		Dumps an MP to a stream.
 */

extern void mp_dump(mp */*x*/, FILE */*fp*/);

/* --- @MP_PARANOID@ --- *
 *
 * Arguments:	@x@ = pointer to MP head
 *
 * Use:		Marks the MP as containing sensitive data which should be
 *		burnt when no longer required.
 */

#define MP_PARANOID(x) ((x)->f |= MPF_BURN)

/* --- @mp_copy@ --- *
 *
 * Arguments:	@mp *d@ = pointer to MP head for destination
 *		@const mp *s@ = pointer to MP head for source
 *
 * Returns:	---
 *
 * Use:		Copies an MP.
 */

extern void mp_copy(mp */*d*/, const mp */*s*/);

/* --- @mp_bits@ --- *
 *
 * Arguments:	@mp *x@ = pointer to MP head
 *
 * Returns:	Length of the number in bits.
 *
 * Use:		Calculates the number of bits required to represent a number.
 *		The number must be normalized.
 */

unsigned long mp_bits(mp */*x*/);

/* --- @mp_octets@ --- *
 *
 * Arguments:	@mp *x@ = pointer to MP head
 *
 * Returns:	Length of number in octets.
 *
 * Use:		Calculates the number of octets required to represent a
 *		number.  The number must be normalized.
 */

extern size_t mp_octets(mp */*x*/);

/*----- Loading and storing as binary data --------------------------------*/

/* --- @mp_storel@ --- *
 *
 * Arguments:	@mp *x@ = pointer to MP head
 *		@octet *p@ = pointer to octet array
 *		@size_t sz@ = size of octet array
 *
 * Returns:	---
 *
 * Use:		Stores an MP in an octet array, least significant octet
 *		first.  High-end octets are silently discarded if there
 *		isn't enough space for them.
 */

extern void mp_storel(mp */*x*/, octet */*p*/, size_t /*sz*/);

/* --- @mp_loadl@ --- *
 *
 * Arguments:	@mp *x@ = pointer to MP head
 *		@const octet *p@ = pointer to octet array
 *		@size_t sz@ = size of octet array
 *
 * Returns:	---
 *
 * Use:		Loads an MP in an octet array, least significant octet
 *		first.
 */

extern void mp_loadl(mp */*x*/, const octet */*p*/, size_t /*sz*/);

/* --- @mp_storeb@ --- *
 *
 * Arguments:	@mp *x@ = pointer to MP head
 *		@octet *p@ = pointer to octet array
 *		@size_t sz@ = size of octet array
 *
 * Returns:	---
 *
 * Use:		Stores an MP in an octet array, most significant octet
 *		first.  High-end octets are silently discarded if there
 *		isn't enough space for them.
 */

extern void mp_storeb(mp */*x*/, octet */*p*/, size_t /*sz*/);

/* --- @mp_loadb@ --- *
 *
 * Arguments:	@mp *x@ = pointer to MP head
 *		@const octet *p@ = pointer to octet array
 *		@size_t sz@ = size of octet array
 *
 * Returns:	---
 *
 * Use:		Loads an MP in an octet array, most significant octet
 *		first.
 */

extern void mp_loadb(mp */*x*/, const octet */*p*/, size_t /*sz*/);

/*----- Iterating through bits --------------------------------------------*/

/* --- @mp_mkbitscan@ --- *
 *
 * Arguments:	@mp_bitscan *sc@ = pointer to bitscan object
 *		@const mp *x@ = pointer to MP head
 *
 * Returns:	---
 *
 * Use:		Initializes a bitscan object.
 */

extern void mp_mkbitscan(mp_bitscan */*sc*/, const mp */*x*/);

/* --- @mp_bstep@ --- *
 *
 * Arguments:	@mp_bitscan *sc@ = pointer to bitscanner object
 *
 * Returns:	Nonzero if there is another bit to read.
 *
 * Use:		Steps on to the next bit, and tells the caller whether one
 *		exists.
 */

extern int mp_bstep(mp_bitscan */*sc*/);

/* --- @mp_bit@ --- *
 *
 * Arguments:	@const mp_bitscan *sc@ = pointer to bitscanner
 *
 * Returns:	Current bit value.
 *
 * Use:		Returns the value of the current bit.
 */

#define MP_BIT(sc) ((sc)->w & 1)

extern int mp_bit(const mp_bitscan */*sc*/);

/*----- Shifting ----------------------------------------------------------*/

/* --- @mp_lsl@ --- *
 *
 * Arguments:	@mp *d@ = pointer to MP head of destination
 *		@const mp *x@ = pointer to MP head of source
 *		@size_t n@ = number of bits to shift
 *
 * Returns:	---
 *
 * Use:		Shifts a number left by a given number of bit positions.
 */

extern void mp_lsl(mp */*d*/, const mp */*x*/, size_t /*n*/);

/* --- @mp_lsr@ --- *
 *
 * Arguments:	@mp *d@ = pointer to MP head of destination
 *		@const mp *x@ = pointer to MP head of source
 *		@size_t n@ = number of bits to shift
 *
 * Returns:	---
 *
 * Use:		Shifts a number right by a given number of bit positions.
 */

extern void mp_lsr(mp */*d*/, const mp */*x*/, size_t /*n*/);

/*----- Unsigned arithmetic -----------------------------------------------*/

/* --- @mp_uadd@ --- *
 *
 * Arguments:	@const mp *d@ = pointers to MP head of destination
 *		@const mp *x, *y@ = pointers to MP heads of operands
 *
 * Returns:	---
 *
 * Use:		Performs unsigned MP addition.
 */

extern void mp_uadd(mp */*d*/, const mp */*x*/, const mp */*y*/);

/* --- @mp_usub@ --- *
 *
 * Arguments:	@const mp *d@ = pointers to MP head of destination
 *		@const mp *x, *y@ = pointers to MP heads of operands
 *
 * Returns:	---
 *
 * Use:		Performs unsigned MP subtraction.
 */

extern void mp_usub(mp */*d*/, const mp */*x*/, const mp */*y*/);

/* --- @mp_ucmp@ --- *
 *
 * Arguments:	@const mp *x, *y@ = pointers to MP heads of operands
 *
 * Returns:	Less than, equal to, or greater than zero.
 *
 * Use:		Performs unsigned MP comparison.
 */

#define MP_UCMP(x, op, y) (mp_ucmp((x), (y)) op 0)

extern int mp_ucmp(const mp */*x*/, const mp */*y*/);

/* --- @mp_umul@ --- *
 *
 * Arguments:	@mp *d@ = pointer to MP head of destination
 *		@const mp *x, *y@ = pointes to MP heads of operands
 *
 * Returns:	---
 *
 * Use:		Performs unsigned MP multiplication.
 */

extern void mp_umul(mp */*d*/, const mp */*x*/, const mp */*y*/);

/* --- @mp_udiv@ --- *
 *
 * Arguments:	@mp *q, *r@ = pointers to MP heads for quotient, remainder
 *		@const mp *x, *y@ = pointers to MP heads for operands
 *
 * Returns:	---
 *
 * Use:		Performs unsigned MP division.
 */

extern void mp_udiv(mp */*q*/, mp */*r*/, const mp */*x*/, const mp */*y*/);

/*----- That's all, folks -------------------------------------------------*/

#ifdef __cplusplus
  }
#endif

#endif
